// Copyright 2017,2018 Lei Ni (nilei81@gmail.com).
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/*
ondisk is an example program for dragonboat's on disk state machine.
*/

package dragonboat

import (
	"context"
	"encoding/json"
	"fmt"
	"os"
	"os/signal"
	"path/filepath"
	"runtime"
	"strings"
	"syscall"
	"time"

	"github.com/lni/dragonboat/v4"
	"github.com/lni/dragonboat/v4/config"
	"github.com/lni/dragonboat/v4/logger"

	"github.com/thanhphu/raftbench/util"
)

// RequestType 定义了请求操作的类型。
type RequestType uint64

// exampleClusterID 是一个预定义的集群 ID，用于演示目的。
const (
	exampleClusterID uint64 = 128
)

// Main 是主入口函数，用于启动一个 dragonboat 节点，并根据传入的参数配置节点的行为。
// 参数:
//   cluster: 集群配置字符串，包含集群成员的信息。
//   nodeID: 当前节点的 ID。
//   addr: 当前节点的地址。
//   join: 是否以加入现有集群的方式启动。
//   test: 测试参数结构体。
func Main(cluster string, nodeID int, addr string, join bool, test util.TestParams) {
	if len(addr) == 0 && nodeID != 1 && nodeID != 2 && nodeID != 3 {
		fmt.Fprintf(os.Stderr, "node id must be 1, 2 or 3 when address is not specified\n")
		os.Exit(1)
	}
	// 解决 macOS 下特定信号处理的问题。
	if runtime.GOOS == "darwin" {
		signal.Ignore(syscall.Signal(0xd))
	}
	initialMembers := make(map[uint64]string)
	if !join {
		// 解析集群配置字符串，获取初始成员列表。
		addresses := strings.Split(strings.Replace(cluster, "http://", "", -1), ",")
		for idx, v := range addresses {
			initialMembers[uint64(idx+1)] = v
		}
	}
	var nodeAddr string
	if len(addr) != 0 {
		nodeAddr = addr
	} else {
		nodeAddr = initialMembers[uint64(nodeID)]
	}
	fmt.Fprintf(os.Stdout, "节点地址: %s\n", nodeAddr)
	// 设置日志级别。
	logger.GetLogger("raft").SetLevel(logger.ERROR)
	logger.GetLogger("rsm").SetLevel(logger.WARNING)
	logger.GetLogger("transport").SetLevel(logger.WARNING)
	logger.GetLogger("grpc").SetLevel(logger.WARNING)
	// 初始化配置对象。
	rc := config.Config{
		ReplicaID:              uint64(nodeID),
		ShardID:                exampleClusterID,
		ElectionRTT:            10,
		HeartbeatRTT:           1,
		CheckQuorum:            true,
		SnapshotEntries:        0,
		DisableAutoCompactions: true,
		MaxInMemLogSize:        0,
		CompactionOverhead:     5,
	}
	// 指定数据目录。
	datadir := filepath.Join(
		fmt.Sprintf("wal-dragonboat-%d", nodeID))
	nhc := config.NodeHostConfig{
		WALDir:         datadir,
		NodeHostDir:    datadir,
		RTTMillisecond: 200,
		RaftAddress:    nodeAddr,
	}
	nh, err := dragonboat.NewNodeHost(nhc)
	if err != nil {
		panic(err)
	}
	if err := nh.StartReplica(initialMembers, join, NewMemKV, rc); err != nil {
		fmt.Fprintf(os.Stderr, "failed to add cluster, %v\n", err)
		os.Exit(1)
	}

	cs := nh.GetNoOPSession(exampleClusterID)

	// 等待分片准备就绪。
	time.Sleep(2 * time.Second)

	ctx, _ := context.WithTimeout(context.Background(), 1000*time.Second)

	// 执行基准测试。
	util.Bench(test, func(k string) bool {
		_, err := nh.SyncRead(ctx, exampleClusterID, k)
		return err == nil
	}, func(k string, v string) bool {
		kv := &KVData{
			Key: k,
			Val: v,
		}
		data, err := json.Marshal(kv)
		if err != nil {
			return false
		}

		_, err = nh.SyncPropose(ctx, cs, data)
		return err == nil
	})
}